import os

from osw.auth import CredentialManager
from osw.core import OSW, model
from osw.ontology import ImportConfig, OntologyImporter, ParserSettings
from osw.wtsite import WtSite

# use credentials from file. if none are found, the user will be prompted to enter them
cm = CredentialManager(
    cred_filepath=os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
    )
)

# create the site object
wtsite = WtSite(WtSite.WtSiteConfig(iri="http://<your-instance>:18081", cred_mngr=cm))
osw = OSW(site=wtsite)

list_of_schemas = [
    "Category:OSW725a3cf5458f4daea86615fcbd0029f8",  # OwlClass
    "Category:OSW6b9ef2784a934b8ab96523366e23e906",  # OwlIndividual
    "Category:OSW57beed5e1294434ba77bb6516e461456",  # EmmoClass
    "Category:Item",
    "Category:ObjectProperty",
    "Category:DataProperty",
    "Category:AnnotationProperty",
]
for i, cat in enumerate(list_of_schemas):
    mode = "append"
    if i == 0:
        mode = "replace"
    osw.fetch_schema(OSW.FetchSchemaParam(schema_title=cat, mode=mode))

ontology_name = "BVCO"

# define ontology metadata
emmo = model.Ontology(
    name="EMMO",
    iri="http://emmo.info/emmo",
    prefix="http://emmo.info/emmo#",
    prefix_name="emmo",
    link="https://github.com/emmo-repo/EMMO",
)
dc = model.Ontology(
    name="DC",
    iri="http://purl.org/dc/elements/1.1",
    prefix="http://purl.org/dc/elements/1.1/",
    prefix_name="dc",
    link="http://purl.org/dc",
)
battinfo = model.Ontology(
    name="EMMO BattINFO",
    iri="http://emmo.info/battery",
    prefix="http://emmo.info/battery#",
    prefix_name="battinfo",
    link="https://github.com/BIG-MAP/BattINFO",
)
electrochemistry = model.Ontology(
    name="EMMO Electrochemistry",
    iri="http://emmo.info/electrochemistry",
    prefix="http://emmo.info/electrochemistry#",
    prefix_name="electrochemistry",
    link="https://github.com/emmo-repo/EMMO",
)
periodictable = model.Ontology(
    name="EMMO Periodic Table",
    iri="http://emmo.info/emmo/domain/periodic-table",
    prefix="http://emmo.info/emmo/domain/periodic-table#",
    prefix_name="periodictable",
    link="https://github.com/emmo-repo/EMMO",
)
gpo = model.Ontology(
    name="GPO",
    iri="https://gpo.ontology.link",
    prefix="https://gpo.ontology.link/",
    prefix_name="gpo",
    link="https://github.com/General-Process-Ontology/ontology",
)
bvco = model.Ontology(
    name="BVCO",
    iri="https://bvco.ontology.link",
    prefix="https://bvco.ontology.link/",
    prefix_name="bvco",
    link="https://github.com/Battery-Value-Chain-Ontology/ontology",
)

import_config = ImportConfig(
    ontology_name=ontology_name,
    ontologies=[emmo, dc, battinfo, electrochemistry, periodictable, gpo, bvco],
    # file="https://raw.githubusercontent.com/BIG-MAP/BattINFO/master/battinfo.ttl",
    file="https://raw.githubusercontent.com/emmo-repo/domain-battery/master/battery.ttl",
    base_class=model.EmmoClass,
    base_class_title="Category:OSW57beed5e1294434ba77bb6516e461456",  # EmmoClass
    dump_files=True,
    dump_path=os.path.dirname(os.path.abspath(__file__)),
    dry_run=False,
)
parser_settings = ParserSettings()
parser_settings.ensure_array.append("elucidation")
parser_settings.ensure_multilang.append("elucidation")

# import ontologies
importer = OntologyImporter(osw=osw, parser_settings=parser_settings)
importer.import_ontology(import_config)



import json
import os
import re
import uuid
from typing import Dict, List, Literal, Optional

import pyld
from pydantic import PrivateAttr
from pydantic.main import ModelMetaclass
from pyld import jsonld
from rdflib import Graph
from typing_extensions import deprecated

from osw.core import OSW, model
from osw.model.static import OswBaseModel
from osw.utils.strings import pascal_case
from osw.utils.wiki import get_namespace
from osw.wtsite import WtSite


class ParserSettings(OswBaseModel):
    """Settings for the ontology parser"""

    flatten_subclassof_restrictions: Optional[bool] = False
    """If True, the parser will directly asign the property and the value class of the restriction to the restricted class
    """
    subclassof_restrictions_keyword: Optional[str] = "restrictions"
    """Keyword to separate the owl restrictions from rdfs subclassof. Only used if flatten_subclassof_restrictions is False."""
    ensure_multilang: Optional[List[str]] = [
        "label",
        "prefLabel",
        "altLabel",
        "comment",
        "description",
    ]
    """List of properties that should be multilang / have a @lang containter. If the property is not multilang, the parser will create a multilang property with "en" as default language."""
    ensure_array: Optional[List[str]] = [
        "rdf_type",
        "label",
        "prefLabel",
        "altLabel",
        "comment",
        "description",
        "subclass_of",
        "restrictions",
    ]
    """List of properties that should be arrays. If the property is not an array, the parser will create an array with the value as single element."""
    map_uuid_iri: Optional[List[str]] = [
        "subclass_of",
        "on_property",
        "some_values_from",
        "all_values_from",
        "domain",
        "range",
        "subproperty_of",
    ]
    """List of properties whose values should be mapped to (OSW) UUIDs. If the value is not a UUID, the parser will create a UUIDv5 from the value with the URL namespace."""
    remove_unnamed: Optional[List[str]] = []  # ['subclass_of'] #, 'equivalentClass']
    """List of properties whose values should be removed if they are unnamed nodes ("_:...")"""


class ImportConfig(OswBaseModel):
    """Configuration for the ontology importer"""

    file: str
    """the path or url to the ontology file"""
    file_format: Literal["xml", "n3", "nt", "trix"] = "n3"
    """the serialization format. for turtle, use 'n3' (see rdflib docs)"""
    ontology_name: str
    """the name of the ontology"""
    ontologies: List[model.Ontology]
    """Ontology metadata, inluding imported ontologies"""
    base_class: ModelMetaclass
    """Base class for the ontology model. For OWL Ontologies, this should be model.OwlClass or a subclass of it."""
    base_class_title: Optional[
        str
    ] = "Category:OSW725a3cf5458f4daea86615fcbd0029f8"  # OwlClass
    """Title of the base class schema"""
    dump_files: Optional[bool] = False
    """If True, the parsed ontology will be dumped to a jsonld file"""
    dump_path: Optional[str] = None
    """Path to the directory where the parsed ontology will be dumped"""
    dry_run: Optional[bool] = False
    """If True, the parsed ontology will not be imported into the wiki"""
    property_naming_policy: Optional[
        Literal["UUID", "label", "prefixed_label"]
    ] = "prefixed_label"
    """Policy for naming properties. If "UUID", the property will be named with a UUIDv5.
    If "label", the property will be named with the label of the property.
    If "prefixed_label", the property will be named with the label of the property prefixed with the label of the ontology."""
    property_naming_prefix_delimiter: Optional[str] = ":"
    """Delimiter for the prefixed_label property naming policy"""

    class Config:
        arbitrary_types_allowed = True


class OntologyImporter(OswBaseModel):
    """Class for importing ontologies into OSW"""

    osw: OSW
    """A OSW instance"""
    parser_settings: Optional[ParserSettings] = ParserSettings()
    """Settings for the ontology parser"""
    import_config: Optional[ImportConfig] = None
    """Configuration for the ontology importer"""

    _context: Optional[Dict] = PrivateAttr()
    _id_dict: Optional[Dict] = PrivateAttr()
    _iri_dict: Optional[Dict] = PrivateAttr()
    _prop_dict: Optional[Dict] = PrivateAttr()
    _g: Optional[Dict] = PrivateAttr()
    _entities: Optional[List] = PrivateAttr()
    _entities_json: Optional[List] = PrivateAttr()

    def import_ontology(self, config: ImportConfig):
        """Imports an ontology into OSW

        Parameters
        ----------
        config
            see ImportConfig

        Returns
        -------
            None
        """

        # overwrite the default document loader to load relative context from the wiki
        def myloader(*args, **kwargs):
            requests_loader = pyld.documentloader.requests.requests_document_loader(
                *args, **kwargs
            )

            def loader(url, options={}):
                if "/wiki/" in url:
                    title = url.replace("/wiki/", "").split("?")[0]
                    page = self.osw.site.get_page(
                        WtSite.GetPageParam(titles=[title])
                    ).pages[0]
                    schema = page.get_slot_content("jsonschema")
                    if isinstance(schema, str):
                        schema = json.loads(schema)
                    doc = {
                        "contentType": "application/json",
                        "contextUrl": None,
                        "documentUrl": url,
                        "document": schema,
                    }
                    return doc

                else:
                    return requests_loader(url, options)

            return loader

        jsonld.set_document_loader(myloader())

        self.import_config = config
        rdf_graph = Graph()
        rdf_graph.parse(self.import_config.file, format=self.import_config.file_format)
        if self.import_config.dump_files:
            rdf_graph.serialize(
                destination=os.path.join(
                    self.import_config.dump_path,
                    f"{self.import_config.ontology_name}.jsonld",
                ),
                format="json-ld",
                auto_compact=True,
            )

        # load the graph into a dict
        self._g = json.loads(rdf_graph.serialize(format="json-ld", auto_compact=True))

        def node_sort(n):
            # sort nodes by type and id
            rank = 10
            if "@type" in n:
                if "Property" in n["@type"]:
                    rank = 0
                if "Class" in n["@type"]:
                    rank = 1
            return (rank, n["@id"])

        self._g["@graph"] = sorted(self._g["@graph"], key=lambda n: node_sort(n))

        # create an id-index of all nodes
        self._id_dict = {}
        for node in self._g["@graph"]:
            if "@id" in node:
                self._id_dict[node["@id"]] = node

        # add a helper context
        self._g["@context"]["_temp_"] = "http://temp.local/"

        # normalize graph
        self._sanitize_graph()

        if self.import_config.dump_files:
            with open(
                os.path.join(
                    self.import_config.dump_path,
                    f"{self.import_config.ontology_name}.mod.jsonld",
                ),
                "w",
                encoding="utf-8",
            ) as f:
                json.dump(self._g, f, indent=4, ensure_ascii=False)

        self._construct_context()
        self._g = jsonld.compact(self._g, self._context)

        # build an iri-index of all nodes
        # ToDo: can we reuse the id-index?
        self._iri_dict = {}
        for node in self._g["@graph"]:
            self._iri_dict[node["iri"]] = node

        self._apply_osw_structure()

        if self.import_config.dump_files:
            with open(
                os.path.join(
                    self.import_config.dump_path,
                    f"{self.import_config.ontology_name}.compacted.jsonld",
                ),
                "w",
                encoding="utf-8",
            ) as f:
                json.dump(self._g, f, indent=4, ensure_ascii=False)

            # currently does not work with relative remote context ("/wiki/..."")
            # rdf_graph.parse(os.path.join(self.import_config.dump_path, f"{self.import_config.ontology_name}.compacted.jsonld"))
            # rdf_graph.serialize(destination=os.path.join(self.import_config.dump_path, f"{self.import_config.ontology_name}.jsonld.ttl"), format="ttl")

        self._create_entities()
        # sort list by type and iri to persist the order across runs
        # self._entities = sorted(self._entities, key=lambda e: (e.type[0], e.iri))

        if self.import_config.dump_files:
            with open(
                os.path.join(
                    self.import_config.dump_path,
                    f"{self.import_config.ontology_name}.entities.jsonld",
                ),
                "w",
                encoding="utf-8",
            ) as f:
                json.dump(self._entities_json, f, indent=4, ensure_ascii=False)

        # import ontologies
        # self._import_ontology(OSW.ImportOntologyParam(ontologies=self.import_config.ontologies, entities=self._entities, dryrun=False))

        if not self.import_config.dry_run:
            self._store_ontologies(
                OntologyImporter.StoreOntologiesParam(
                    ontologies=self.import_config.ontologies,
                    entities=self._entities,
                )
            )

    def _sanitize_graph(self) -> None:
        ps = self.parser_settings
        # sanitize types of type object, see: https://github.com/RDFLib/rdflib/issues/1950
        for node in self._g["@graph"]:
            key = "@type"
            if key in node:
                # normalize to list
                if not isinstance(node[key], list):
                    node[key] = [node[key]]
                types = []
                for i, val in enumerate(node[key]):
                    if isinstance(val, dict):
                        if "@id" in val:
                            types.append(val["@id"])
                    else:
                        types.append(val)
                node[key] = types
            # remove 'www.' from @id
            node["@id"] = node["@id"].replace("www.", "")
            if "rdfs:subClassOf" in node:
                if not isinstance(node["rdfs:subClassOf"], list):
                    node["rdfs:subClassOf"] = [node["rdfs:subClassOf"]]
                for i, val in enumerate(node["rdfs:subClassOf"]):
                    id = val["@id"]
                    if id in self._id_dict:
                        superclass = self._id_dict[id]
                        if (
                            "@type" in superclass
                            and superclass["@type"] == "owl:Restriction"
                            and "owl:onProperty" in superclass
                            and "owl:someValuesFrom" in superclass
                        ):
                            if ps.flatten_subclassof_restrictions:
                                if superclass["owl:onProperty"]["@id"] not in node:
                                    node[superclass["owl:onProperty"]["@id"]] = []
                                node[superclass["owl:onProperty"]["@id"]].append(
                                    superclass["owl:someValuesFrom"]["@id"]
                                )
                            else:
                                if (
                                    "_temp_:" + ps.subclassof_restrictions_keyword
                                    not in node
                                ):
                                    node[
                                        "_temp_:" + ps.subclassof_restrictions_keyword
                                    ] = []
                                node[
                                    "_temp_:" + ps.subclassof_restrictions_keyword
                                ].append(superclass)
                            node["rdfs:subClassOf"].pop(i)
                            if id in self._g["@graph"]:
                                self._delete_node(self._g["@graph"], id)

    def _delete_node(self, id: str, key: str = "@id") -> bool:
        """Delete a node from a graph by id

        Parameters
        ----------
        graph
            the graph (list of nodes) to delete from
        id
            the id of the node to delete
        key, optional
            the id key, by default '@id'

        Returns
        -------
            True if the node was deleted, False otherwise
        """
        graph = self._g["@graph"]
        for i, node in enumerate(graph):
            if node[key] == id:
                del graph[i]
                return True
        return False

    def _construct_context(self):
        self._context = {
            "_temp_": "http://temp.local/",
            "owl": "http://www.w3.org/2002/07/owl#",
            "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "xsd": "http://www.w3.org/2001/XMLSchema#",
            "skos": "http://www.w3.org/2004/02/skos/core#",
            "dc": "http://purl.org/dc/terms/",
            # "emmo": "http://emmo.info/emmo#", #keep values with full iri
            "iri": {"@id": "@id"},
            "rdf_type": {"@id": "@type"},
            "label": {"@id": "rdfs:label"},
            # "label": {"@id": "skos:prefLabel"},
            "altLabel": {"@id": "skos:altLabel"},
            "text": {"@id": "@value"},
            "lang": {"@id": "@language"},
            # "subClassOf": {"@id": "rdfs:subClassOf", "@type": "@id"},
            "subclass_of": {"@id": "rdfs:subClassOf", "@type": "@id"},
            "source": "dc:source",
            "disjointUnionOf": "owl:disjointUnionOf",
            "disjointWith": "owl:disjointWith",
            "equivalentClass": "owl:equivalentClass",
            "unionOf": {"@id": "owl:unionOf", "@container": "@list", "@type": "@id"},
            "comment": "rdfs:comment",
            "isDefinedBy": "rdfs:isDefinedBy",
            "seeAlso": "rdfs:seeAlso",
            # shorten properties
            "description": "http://example.com/R7k3ssL7gUxsfWuVsXWDXYF",
            # "ObjectPropertyA": "http://example.com/REh2qNSARmKpPuwrJmr5Pu",
            "Item": "https://wiki-dev.open-semantic-lab.org/wiki",
            # "Item:OSW4db96fb7e9e0466d942bf6f17bfdc145": "emmo:EMMO_4db96fb7_e9e0_466d_942b_f6f17bfdc145",
            # "uuid": "wiki:HasUuid",
            # "name": "wiki:HasName",
            # "restrictions": {"@id": "rdfs:subClassOf", "@type": "@id"},
            "restrictions": {"@id": "_temp_:restrictions", "@type": "@id"},
            "on_property": {"@id": "owl:onProperty", "@type": "@id"},
            "some_values_from": {"@id": "owl:someValuesFrom", "@type": "@id"},
            "domain": {"@id": "rdfs:domain", "@type": "@id"},
            "range": {"@id": "rdfs:range", "@type": "@id"},
            "subproperty_of": {"@id": "rdfs:subPropertyOf", "@type": "@id"},
            # "example": "http://example.com/",
        }

        # base_class_page = self.osw.site.get_page(WtSite.GetPageParam(
        #    titles=[self.import_config.base_class_title]
        # )).pages[0]
        # schema = base_class_page.get_slot_content("jsonschema")
        # if (isinstance(schema, str)):
        #    schema = json.loads(schema)
        # self._context = schema["@context"]

        self._context = [
            f"/wiki/{self.import_config.base_class_title}?action=raw&slot=jsonschema&format=json",
            {
                "_temp_": "http://temp.local/",
                "restrictions": {"@id": "_temp_:restrictions", "@type": "@id"},
            },
        ]

    def _get_uuid_from_iri(self, iri):
        node_name = iri.split("/")[-1]
        if "#" in node_name:
            node_name = node_name.split("#")[-1]
        try:
            # try to get last 32 hex digits and convert it to uuid
            # e. g. http://emmo.info/emmo/EMMO_4db96fb7_e9e0_466d_942b_f6f17bfdc145
            node_uuid = uuid.UUID(re.sub(r"[^A-Fa-f0-9]", "", node_name)[-32:])
        except Exception:
            node_uuid = uuid.uuid5(uuid.NAMESPACE_URL, iri)
        return node_uuid

    def _get_page_name(self, node):
        ns = ""
        title = ""

        node_uuid = self._get_uuid_from_iri(node["iri"])

        if "owl:Class" in node["rdf_type"]:
            ns = "Category"
            title = OSW.get_osw_id(node_uuid)
        elif (
            "owl:ObjectProperty" in node["rdf_type"]
            or "owl:DatatypeProperty" in node["rdf_type"]
            or "owl:AnnotationProperty" in node["rdf_type"]
        ):
            ns = "Property"
            if self.import_config.property_naming_policy == "label":
                title = node["name"]
            elif self.import_config.property_naming_policy == "prefixed_label":
                prefix = None
                for onto in self.import_config.ontologies:
                    if (
                        onto.iri in node["iri"]
                        or onto.prefix_name
                        + self.import_config.property_naming_prefix_delimiter
                        in node["iri"]
                    ):
                        prefix = onto.prefix_name
                        break

                if prefix:
                    name = None
                    if "name" in node:
                        name = node["name"]
                    if ":" in node["iri"]:
                        name = node["iri"].split(":")[-1]
                    if name:
                        title = f"{prefix}{self.import_config.property_naming_prefix_delimiter}{name}"
                    else:
                        raise ValueError(
                            f"Could not find name for property {node['iri']}"
                        )
                else:
                    raise ValueError(
                        f"Could not find prefix for property {node['iri']}"
                    )
            if self.import_config.property_naming_policy == "uuid":
                title = OSW.get_osw_id(node_uuid)
        elif "owl:NamedIndividual" in node["rdf_type"]:
            ns = "Item"
            title = OSW.get_osw_id(node_uuid)
        else:  # e. g. owl:Axiom
            return None

        return f"{ns}:{title}"

    def _map_iri_to_osw(self, iri):
        result = None
        if iri in self._iri_dict:
            node = self._iri_dict[iri]
            result = self._get_page_name(node)
        if not result:
            result = iri
        return result

    def _apply_osw_structure(self):
        ps = self.parser_settings
        for node in self._g["@graph"]:
            for key in ps.ensure_multilang:
                if key in node:
                    if isinstance(node[key], str):
                        node[key] = {"text": node[key], "lang": "en"}
                    elif "text" in node[key] and "lang" not in node[key]:
                        node[key]["lang"] = "en"
                    elif isinstance(node[key], list):
                        for i, val in enumerate(node[key]):
                            if isinstance(node[key][i], str):
                                node[key][i] = {"text": node[key][i], "lang": "en"}
                            elif "text" in node[key][i] and "lang" not in node[key][i]:
                                node[key][i]["lang"] = "en"
            for key in ps.ensure_array:
                if key in node and not isinstance(node[key], list):
                    node[key] = [node[key]]
            for key in ps.remove_unnamed:
                if key in node:
                    if isinstance(node[key], list):
                        node[key] = [
                            value for value in node[key] if not value.startswith("_:")
                        ]
                    elif isinstance(node[key], str) and node[key].startswith("_:"):
                        del node[key]

            if "rdf_type" in node and any(
                x in node["rdf_type"]
                for x in [
                    "owl:Class",
                    "owl:ObjectProperty",
                    "owl:DatatypeProperty",
                    "owl:AnnotationProperty",
                    "owl:NamedIndividual",
                ]
            ):
                node["uuid"] = str(self._get_uuid_from_iri(node["iri"]))

                # name = ""
                if "prefLabel" in node:
                    # if isinstance(node["prefLabel"], str): node["prefLabel"] = {"text": node["prefLabel"], "lang": "en"}
                    # if isinstance(node["prefLabel"], list): node["name"] = node["prefLabel"][0]["text"]
                    # else: node["name"] = node["prefLabel"]["text"]
                    node["name"] = node["prefLabel"][0]["text"]
                elif "label" in node:
                    node["name"] = node["label"][0]["text"]
                else:
                    print("No label: ", node["iri"])
                if "name" in node:
                    node["name"] = pascal_case(node["name"])

                # update iri dict
                self._iri_dict[node["iri"]] = node

                namespace = self._get_page_name(node).split(":")[0]
                title = self._get_page_name(node).replace(namespace + ":", "")
                node["meta"] = {
                    "wiki_page": {
                        "title": title,
                        "namespace": namespace,
                    }
                }

                for onto in self.import_config.ontologies:
                    if (
                        onto.prefix in node["iri"]
                        or onto.prefix_name + ":" in node["iri"]
                    ):
                        node["imported_from"] = (
                            onto.prefix_name
                            + ":"
                            + node["iri"]
                            .replace(onto.prefix, "")
                            .replace(onto.prefix_name + ":", "")
                        )
                        break

        self._map_iris(self._g["@graph"])

        # build an inverted property_dict
        self._prop_dict = {}
        for node in self._g["@graph"]:
            self._iri_dict[node["iri"]] = node
        for node in self._g["@graph"]:
            for key in node:
                if ":" in key and key not in self._prop_dict:
                    label = ""
                    if key in self._iri_dict and "label" in self._iri_dict[key]:
                        label = self._iri_dict[key]["label"][0]["text"]
                    self._prop_dict[key] = label
        # print the inverted property_dict
        for key, value in self._prop_dict.items():
            print(f'"{value}": "{key}"')

    def _map_iris(self, node_array):
        if not isinstance(node_array, list):
            node_array = [node_array]
        for node in node_array:
            if isinstance(node, dict):
                for key in node:
                    if key in self.parser_settings.map_uuid_iri:
                        if isinstance(node[key], list):
                            for i, val in enumerate(node[key]):
                                node[key][i] = self._map_iri_to_osw(node[key][i])
                        if isinstance(node[key], str):
                            node[key] = self._map_iri_to_osw(node[key])
                    if key == "restrictions":
                        print(key)
                    if isinstance(node[key], dict) or isinstance(node[key], list):
                        self._map_iris(node[key])

    def _create_entities(self):
        # create OSW entities
        limit = 3000  # choose a smaller number for tests
        counter = 0
        self._entities = []
        self._entities_json = []
        for index, node in enumerate(self._g["@graph"]):
            # restore class restrictions
            # if "_restrictions" in node:
            #    node["restrictions"] == node["_restrictions"]

            if counter < limit:
                if "rdf_type" in node and "label" in node:
                    e = None
                    if "owl:Class" in node["rdf_type"]:
                        e = self.import_config.base_class(**node)

                    if "owl:ObjectProperty" in node["rdf_type"]:
                        e = model.ObjectProperty(**node)

                    if "owl:DatatypeProperty" in node["rdf_type"]:
                        e = model.DataProperty(**node)

                    if "owl:AnnotationProperty" in node["rdf_type"]:
                        e = model.AnnotationProperty(**node)

                    if "owl:NamedIndividual" in node["rdf_type"]:
                        e = model.OwlIndividual(**node)

                    if e:
                        self._entities.append(e)
                        self._entities_json.append(
                            json.loads(e.json(exclude_none=True))
                        )
                        counter += 1

    class StoreOntologyParam(model.OswBaseModel):
        ontology: model.Ontology
        entities: List[model.OswBaseModel]

    def _store_ontology(self, param: StoreOntologyParam):
        import_page = self.osw.site.get_page(
            WtSite.GetPageParam(
                titles=["MediaWiki:Smw_import_" + param.ontology.prefix_name]
            )
        ).pages[0]
        text = (
            f"{param.ontology.prefix} | [{param.ontology.link} {param.ontology.name}]"
        )
        for e in param.entities:
            namespace = get_namespace(e)

            # see https://www.semantic-mediawiki.org/wiki/Help:Import_vocabulary
            if namespace == "Category":
                smw_import_type = "Category"
            elif namespace == "Property":
                smw_import_type = "Type:" + e.cast(model.Property).property_type
            else:
                smw_import_type = "Item"
            iri = None
            if hasattr(e, "iri"):
                iri = e.iri
            if hasattr(e, "uri"):
                iri = e.uri
            if iri is not None:
                text += f"\n {iri.replace(param.ontology.prefix, '').replace(param.ontology.prefix_name + ':', '')}|{smw_import_type}"
            else:
                print("Error: Entity has not iri/uri property")

        import_page.set_slot_content("main", text)
        import_page.edit("import ontology")

        self.osw.store_entity(
            OSW.StoreEntityParam(namespace="Category", entities=param.entities)
        )

    @deprecated("use ontology.OntologyImporter.StoreOntologiesParam instead")
    class StoreOntologiesParam(model.OswBaseModel):
        entities: Optional[List[model.OswBaseModel]]
        """if we use model.Entity here, all instances are casted to model.Entity
        see: https://stackoverflow.com/questions/67366187/how-to-make-a-pydantic-field-accept-subclasses-using-type
        """
        # classes: Optional[List[model.Entity]]  # goes to NS Category
        # properties: Optional[List[model.Entity]]  # goes to NS Property
        # individuals: Optional[List[model.Entity]]  # goes to NS Item
        ontologies: Optional[List[model.Ontology]]
        dryrun: Optional[bool] = False

    def _store_ontologies(self, param: StoreOntologiesParam = None):
        entities = self._entities
        dry_run = self.import_config.dry_run
        ontologies = self.import_config.ontologies
        if param:
            entities = param.entities
            dry_run = param.dryrun
            ontologies = param.ontologies

        print(f"{len(entities)} classes to import")
        prefix_dict = {}
        for e in entities:
            if "//" in e.iri:
                if "http://purl.obolibrary.org/obo/" in e.iri:
                    # special handling for OBO ontologies
                    # ontologies are distinguished by the first part of the postfix
                    # e. g. http://purl.obolibrary.org/obo/BFO_0000050
                    postfix = e.iri.replace("http://purl.obolibrary.org/obo/", "")
                    key = (
                        "http://purl.obolibrary.org/obo/" + postfix.split("_")[0] + "_"
                    )
                elif "#" in e.iri:
                    key = e.iri.split("#")[0] + "#"
                else:
                    key = e.iri.replace(e.iri.split("/")[-1], "")
            else:
                key = e.iri.split(":")[0]  # e. g. "obo:OBO_0000050"
            if key not in prefix_dict:
                prefix_dict[key] = []
            prefix_dict[key].append(e)

        for prefix in prefix_dict.keys():
            onto = None
            for o in ontologies:
                if o.prefix == prefix or o.prefix_name == prefix:
                    onto = o
            if onto is None:
                print(f"Error: No ontology defined for prefix {prefix}")
            else:
                if not dry_run:
                    self._store_ontology(
                        OntologyImporter.StoreOntologyParam(
                            ontology=onto, entities=prefix_dict[prefix]
                        )
                    )



import os

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.core import OSW
from osw.wtsite import WtSite

# create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)
osw = OSW(site=wtsite)

wtsite.enable_cache()  # skip downloads on second request

# osw.fetch_schema() #this will load the current entity schema from the OSW instance. You may have to re-run the script to get the updated schema extension. Requires 'pip install datamodel-code-generator'
if True:
    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title="Category:OSW02590972aeba46d7864ed492c0c11384",  # Host
            mode="replace",
        )
    )
    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title="Category:OSWacdb001c926c46b998af3e645d36b13f",
            # DatabaseServer
            mode="append",
        )
    )
    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title="Category:OSW51ad0d1716254c77a2b7a03217f23b43",  # Database
            mode="append",
        )
    )
    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title="Category:OSW33c7c3a4076a4a58b20d818c162e84e6",  # DatabaseType
            mode="append",
        )
    )


# make sure to import controllers after updating the model (ignore linter warning)
import osw.controller as controller  # noqa: E402

# load database definition
db = osw.load_entity("Item:OSWb8cc7705e17c47b19331fdb045bfbca8")  # postgres
db = db.cast(model.Database)

# cast into controller and execute function
db = db.cast(controller.DatabaseController)
db.connect(
    controller.DatabaseController.ConnectionConfig(
        cm=CredentialManager(cred_filepath=pwd_file_path), osw=osw
    )
)
db.execute("SELECT * FROM public.sensors ORDER BY sensor_id ASC")



from typing import Union

import osw.model.entity as model


def get_namespace(entity: model.Entity) -> Union[str, None]:
    """determines the wiki namespace based on the entity's type/class

    parameters:
        entity: the entity to determine the namespace for

    returns:
        the namespace as a string or None if the namespace could not be determined
    """
    namespace = None
    #  (model classes may not exist => try except)
    if namespace is None:
        try:
            if issubclass(entity, model.Entity):
                namespace = "Category"
        except (TypeError, AttributeError):
            pass
    if namespace is None:
        try:
            if isinstance(entity, model.Item):
                namespace = "Item"
        except AttributeError:
            pass
    if namespace is None:
        try:
            if isinstance(entity, model.Category):
                namespace = "Category"
        except AttributeError:
            pass
    if namespace is None:
        try:
            if isinstance(entity, model.Property):
                namespace = "Property"
        except AttributeError:
            pass
    if namespace is None:
        try:
            if isinstance(entity, model.WikiFile):
                namespace = "File"
        except AttributeError:
            pass

    return namespace



"""
This module is to be imported in the dynamically created and updated entity.py module.
"""
from typing import Type, TypeVar, Union

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class OswBaseModel(BaseModel):
    def full_dict(self, **kwargs):  # extent BaseClass export function
        d = super().dict(**kwargs)
        for key in ("_osl_template", "_osl_footer"):
            if hasattr(self, key):
                d[key] = getattr(self, key)
                # Include selected private properties. note: private properties are not
                #  considered as discriminator
        return d

    def cast(self, cls: Union[Type[T], type]) -> T:
        return cls(**self.dict())

    def cast_none_to_default(self, cls: Union[Type[T], type]) -> T:
        """Casting self into target class. If the passed attribute is None or solely
        includes None values, the attribute is not passed to the instance of the
        target class, which will then fall back to the default."""

        def test_if_empty_list_or_none(obj) -> bool:
            if obj is None:
                return True
            elif isinstance(obj, list):
                if len(obj) == 0:
                    return True
                elif len([item for item in obj if item is not None]) == 0:
                    return True
            return False

        return cls(
            **{
                k: v
                for k, v in self.dict().items()
                if not test_if_empty_list_or_none(v)
            }
        )


class Ontology(OswBaseModel):
    iri: str
    prefix: str
    name: str
    prefix_name: str
    link: str



from typing import Optional, Union

from sqlalchemy import URL, create_engine
from sqlalchemy import text as sql_text
from sqlalchemy.engine import Engine

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.core import OSW
from osw.model.static import OswBaseModel


class DatabaseController(model.Database):
    """
    Provides a sqlalchemy engine for the specified database
    """

    osw: Optional[OSW]
    """ an OSW instance to fetch related resources (host, server, etc.)"""
    cm: Optional[CredentialManager]
    """ CredentialManager to login to the database"""
    engine: Optional[Engine]
    """ the internal sqlalchemy engine """

    class Config:
        arbitrary_types_allowed = True

    class ConnectionString(OswBaseModel):
        """the sqlalchemy database connection string"""

        dialect: str
        """ sql subtype, e. g. postgres"""
        username: str
        """ username for the login """
        password: str
        """ password for the login """
        host: str
        """ the database host (ip address or domain) """
        port: int
        """ the database port """
        database: str
        """ the database name / identifier """
        driver: Optional[str]
        """ specific driver, e. g. psycopg2 """

        def __str__(self):
            """generates the string representation

            Returns
            -------
                database connection string
            """
            prefix = f"{self.dialect}"
            if self.driver:
                prefix += f"+{self.driver}"
            url = URL.create(
                prefix,
                username=self.username,
                password=self.password,
                host=self.host,
                port=self.port,
                database=self.database,
            )
            return url.render_as_string(hide_password=False)

    class ConnectionConfig(OswBaseModel):
        """Database connection configuration"""

        osw: OSW
        """ OSW instance to fetch related resources (host, server, etc.) """
        cm: Union[CredentialManager.Credential, CredentialManager]
        """ CredentialManager or direct Credential to login to the database"""

    def connect(self, config: ConnectionConfig):
        """Initializes the connection to the database by creating a sqlalchemy engine

        Parameters
        ----------
        config
            see ConnectionConfig
        """
        self.osw = config.osw
        self.cm = config.cm

        server_title = self.osw.site.semantic_search(
            f"[[-HasDbServer::Item:{self.osw.get_osw_id(self.uuid)}]]"
        )
        server = self.osw.load_entity(server_title[0]).cast(model.DatabaseServer)

        host_title = self.osw.site.semantic_search(
            f"[[-HasHost::Item:{self.osw.get_osw_id(server.uuid)}]]"
        )
        host = self.osw.load_entity(host_title[0]).cast(model.Host)

        dbtype_title = self.osw.site.semantic_search(
            f"[[-HasDbType::Item:{self.osw.get_osw_id(server.uuid)}]]"
        )
        dbtype = self.osw.load_entity(dbtype_title[0]).cast(model.DatabaseType)

        if type(self.cm) is CredentialManager.Credential:
            db_server_cred = self.cm
        else:
            db_server_cred = config.cm.get_credential(
                CredentialManager.CredentialConfig(
                    iri=f"{host.network_domain[0]}:{server.network_port[0]}",
                    fallback=CredentialManager.CredentialFallback.ask,
                )
            )

        cstr = DatabaseController.ConnectionString(
            dialect=dbtype.connection_str_dialect,
            driver=dbtype.connection_str_driver,
            username=db_server_cred.username,
            password=db_server_cred.password,
            host=host.network_domain[0],
            port=server.network_port[0],
            database=self.name,
        )

        self.engine = create_engine(str(cstr))

    def execute(self, sql: str):
        """Executes a plain sql string

        Parameters
        ----------
        sql
            the sql string
        """
        with self.engine.connect() as conn:
            result_set = conn.execute(sql_text(sql))
            for r in result_set:
                print(r)



import os

import osw.model.entity as model
from osw.core import OSW
from osw.wtsite import WtSite

# create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)
osw = OSW(site=wtsite)

title = "Item:OSW7d7193567ea14e4e89b74de88983b718"
# title = "Item:OSWe02213b6c4664d04834355dc8eb08b99"
entity = osw.load_entity(title)
print(entity.__class__)
print(entity.label[0].text)  # we can access any attribute of model.Entity

hardware_entity = entity.cast(model.Hardware)  # explicit cast to model.Hardware
print(hardware_entity.manufacturer)  # we can access now any attribute of model.Entity

print(entity.manufacturer)

print(entity.json(exclude_none=True))



from __future__ import annotations

import importlib
import json
import os
import platform
import re
import sys
from enum import Enum
from typing import List, Optional, Union
from uuid import UUID

from jsonpath_ng.ext import parse
from pydantic import BaseModel, Field, create_model
from pydantic.main import ModelMetaclass

import osw.model.entity as model
from osw.model.static import OswBaseModel
from osw.utils.util import parallelize
from osw.utils.wiki import get_namespace
from osw.wiki_tools import SearchParam
from osw.wtsite import WtSite


class OswClassMetaclass(ModelMetaclass):
    def __new__(cls, name, bases, dic, osl_template, osl_footer_template):
        base_footer_cls = type(
            dic["__qualname__"] + "Footer",
            (BaseModel,),
            {
                "__annotations__": {"osl_template": str},
                "osl_template": Field(
                    default=osl_footer_template,
                    title=dic["__qualname__"] + "FooterTemplate",
                ),
            },
        )
        if "__annotations__" not in dic:
            dic["__annotations__"] = {}
        dic["__annotations__"]["osl_template"] = str
        dic["osl_template"] = Field(
            default=osl_template, title=dic["__qualname__"] + "Template"
        )
        dic["__annotations__"]["osl_footer"] = base_footer_cls
        dic["osl_footer"] = Field(
            default={"osl_template": osl_footer_template},
            title=dic["__qualname__"] + "Footer",
        )
        new_cls = super().__new__(cls, name, bases, dic)
        return new_cls


class OSW(BaseModel):
    """Bundles core functionalities of OpenSemanticWorld (OSW)"""

    uuid: str = "2ea5b605-c91f-4e5a-9559-3dff79fdd4a5"
    _protected_keywords = (
        "_osl_template",
        "_osl_footer",
    )  # private properties included in model export

    class Config:
        arbitrary_types_allowed = True  # neccessary to allow e.g. np.array as type

    site: WtSite

    @staticmethod
    def get_osw_id(uuid: uuid) -> str:
        """Generates a OSW-ID based on the given uuid by prefixing "OSW" and removing
        all '-' from the uuid-string

        Parameters
        ----------
        uuid
            uuid object, e. g. UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")

        Returns
        -------
            OSW-ID string, e. g. OSW2ea5b605c91f4e5a95593dff79fdd4a5
        """
        return "OSW" + str(uuid).replace("-", "")

    @staticmethod
    def get_uuid(osw_id) -> uuid:
        """Returns the uuid for a given OSW-ID

        Parameters
        ----------
        osw_id
            OSW-ID string, e. g. OSW2ea5b605c91f4e5a95593dff79fdd4a5

        Returns
        -------
            uuid object, e. g. UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")
        """
        return UUID(osw_id.replace("OSW", ""))

    class SchemaRegistration(BaseModel):
        """
        dataclass param of register_schema()

        Attributes
        ----------
        model_cls:
            the model class
        schema_name:
            the name of the schema
        schema_bases:
            list of base schemas (referenced by allOf)
        """

        class Config:
            arbitrary_types_allowed = True  # allow any class as type

        model_cls: ModelMetaclass
        schema_uuid = str  # Optional[str] = model_cls.__uuid__
        schema_name: str  # Optional[str] = model_cls.__name__
        schema_bases: List[str] = ["Category:Item"]

    def register_schema(self, schema_registration: SchemaRegistration):
        """Registers a new or updated schema in OSW by creating the corresponding
        category page.

        Parameters
        ----------
        schema_registration
            see SchemaRegistration
        """
        entity = schema_registration.model_cls

        jsondata = {}
        jsondata["uuid"] = schema_registration.schema_uuid
        jsondata["label"] = {"text": schema_registration.schema_name, "lang": "en"}
        jsondata["subclass_of"] = schema_registration.schema_bases

        if issubclass(entity, BaseModel):
            entity_title = "Category:" + OSW.get_osw_id(schema_registration.schema_uuid)
            page = self.site.get_page(WtSite.GetPageParam(titles=[entity_title])).pages[
                0
            ]

            page.set_slot_content("jsondata", jsondata)

            # entity = ModelMetaclass(entity.__name__, (BaseModel,), dict(entity.__dict__)) #strips base classes but fiels are already importet
            schema = json.loads(
                entity.schema_json(indent=4).replace("$ref", "dollarref")
            )

            jsonpath_expr = parse("$..allOf")
            # replace local definitions (#/definitions/...) with embedded definitions to prevent resolve errors in json-editor
            for match in jsonpath_expr.find(schema):
                result_array = []
                for subschema in match.value:
                    # pprint(subschema)
                    value = subschema["dollarref"]
                    if value.startswith("#"):
                        definition_jsonpath_expr = parse(
                            value.replace("#", "$").replace("/", ".")
                        )
                        for def_match in definition_jsonpath_expr.find(schema):
                            # pprint(def_match.value)
                            result_array.append(def_match.value)
                    else:
                        result_array.append(subschema)
                match.full_path.update_or_create(schema, result_array)
            if "definitions" in schema:
                del schema["definitions"]

            if "allOf" not in schema:
                schema["allOf"] = []
            for base in schema_registration.schema_bases:
                schema["allOf"].append(
                    {"$ref": f"/wiki/{base}?action=raw&slot=jsonschema"}
                )

            page.set_slot_content("jsonschema", schema)
        else:
            print("Error: Unsupported entity type")
            return

        page.edit()
        print("Entity stored at " + page.get_url())

    class SchemaUnregistration(BaseModel):
        """
        dataclass param of register_schema()

        Attributes
        ----------
        model_cls:
            the model class
        schema_name:
            the name of the schema
        schema_bases:
            list of base schemas (referenced by allOf)
        """

        class Config:
            arbitrary_types_allowed = True  # allow any class as type

        model_cls: Optional[ModelMetaclass]
        model_uuid: Optional[str]
        comment: Optional[str]

    def unregister_schema(self, schema_unregistration: SchemaUnregistration):
        """deletes the corresponding category page

        Parameters
        ----------
        schema_unregistration
            see SchemaUnregistration
        """
        uuid = ""
        if schema_unregistration.model_uuid:
            uuid = schema_unregistration.model_uuid
        elif (
            not uuid
            and schema_unregistration.model_cls
            and issubclass(schema_unregistration.model_cls, BaseModel)
        ):
            uuid = schema_unregistration.model_cls.__uuid__
        else:
            print("Error: Neither model nor model id provided")

        entity_title = "Category:" + OSW.get_osw_id(uuid)
        page = self.site.get_page(WtSite.GetPageParam(titles=[entity_title])).pages[0]
        page.delete(schema_unregistration.comment)

    class FetchSchemaMode(Enum):
        """Modes of the FetchSchemaParam class

        Attributes
        ----------
        append:
            append to the current model
        replace:
            replace the current model
        """

        append = "append"  # append to the current model
        replace = "replace"  # replace the current model

    class FetchSchemaParam(BaseModel):
        """_summary_

        Attributes
        ----------
        schema_title:
            the title (wiki page name) of the schema (default: Category:Item)
        root:
            marks the root iteration for a recursive fetch (internal param,
            default: True)
        mode:
            append or replace (default) current schema, see FetchSchemaMode
        """

        schema_title: Optional[str] = "Category:Item"
        root: Optional[bool] = True
        mode: Optional[
            str
        ] = "replace"  # type 'FetchSchemaMode' requires: 'from __future__ import annotations'

    def fetch_schema(self, fetchSchemaParam: FetchSchemaParam = None) -> None:
        """Loads the given schema from the OSW instance and autogenerates python
        datasclasses within osw.model.entity from it

        Parameters
        ----------
        fetchSchemaParam
            See FetchSchemaParam, by default None
        """
        site_cache_state = self.site.get_cache_enabled()
        self.site.enable_cache()
        if fetchSchemaParam is None:
            fetchSchemaParam = OSW.FetchSchemaParam()
        schema_title = fetchSchemaParam.schema_title
        root = fetchSchemaParam.root
        schema_name = schema_title.split(":")[-1]
        page = self.site.get_page(WtSite.GetPageParam(titles=[schema_title])).pages[0]
        if not page.exists:
            print(f"Error: Page {schema_title} does not exist")
            return
        if schema_title.startswith("Category:"):
            schema_str = json.dumps(page.get_slot_content("jsonschema"))
        else:
            schema_str = page.get_content()
        schema = json.loads(
            schema_str.replace("$ref", "dollarref")
        )  # '$' is a special char for root object in jsonpath
        print(f"Fetch {schema_title}")

        jsonpath_expr = parse("$..dollarref")
        for match in jsonpath_expr.find(schema):
            # value = "https://" + self.site._site.host + match.value
            if match.value.startswith("#"):
                continue  # skip self references
            ref_schema_title = match.value.replace("/wiki/", "").split("?")[0]
            ref_schema_name = ref_schema_title.split(":")[-1] + ".json"
            value = ""
            for i in range(0, schema_name.count("/")):
                value += "../"  # created relative path to top-level schema dir
            value += ref_schema_name  # create a reference to a local file
            match.full_path.update_or_create(schema, value)
            # print(f"replace {match.value} with {value}")
            if (
                ref_schema_title != schema_title
            ):  # prevent recursion in case of self references
                self.fetch_schema(
                    OSW.FetchSchemaParam(schema_title=ref_schema_title, root=False)
                )  # resolve references recursive

        model_dir_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "model"
        )  # src/model
        schema_path = os.path.join(model_dir_path, schema_name + ".json")
        os.makedirs(os.path.dirname(schema_path), exist_ok=True)
        with open(schema_path, "w", encoding="utf-8") as f:
            schema_str = json.dumps(schema, ensure_ascii=False, indent=4).replace(
                "dollarref", "$ref"
            )
            # print(schema_str)
            f.write(schema_str)

        # result_model_path = schema_path.replace(".json", ".py")
        result_model_path = os.path.join(model_dir_path, "entity.py")
        temp_model_path = os.path.join(model_dir_path, "temp.py")
        if root:
            exec_name = "datamodel-codegen"
            if platform.system() == "Windows":
                exec_name += ".exe"
            exec_path = os.path.join(
                os.path.dirname(os.path.abspath(sys.executable)), exec_name
            )
            if not os.path.isfile(exec_path):
                exec_path = os.path.join(
                    os.path.dirname(os.path.abspath(sys.executable)),
                    "Scripts",
                    exec_name,
                )
            if not os.path.isfile(exec_path):
                print("Error: datamodel-codegen not found")
                return
            os.system(
                f"{exec_path}  \
                --input {schema_path} \
                --input-file-type jsonschema \
                --output {temp_model_path} \
                --base-class osw.model.static.OswBaseModel \
                --use-default \
                --enum-field-as-literal all \
                --use-title-as-name \
                --use-schema-description \
                --use-field-description \
                --encoding utf-8 \
                --use-double-quotes \
                --collapse-root-models \
                --reuse-model \
            "
            )
            # see https://koxudaxi.github.io/datamodel-code-generator/
            # --base-class OswBaseModel: use a custom base class
            # --custom-template-dir src/model/template_data/
            # --extra-template-data src/model/template_data/extra.json
            # --use-default: Use default value even if a field is required
            # --enum-field-as-literal all: prevent 'value is not a valid enumeration member' errors after schema reloading
            # --use-schema-description: Use schema description to populate class docstring
            # --use-field-description: Use schema description to populate field docstring
            # --use-title-as-name: use titles as class names of models, e. g. for the footer templates
            # --collapse-root-models: Models generated with a root-type field will be merged
            # into the models using that root-type model, e. g. for Entity.statements
            # --reuse-model: Re-use models on the field when a module has the model with the same content

            # this is dirty, but required for autocompletion: https://stackoverflow.com/questions/62884543/pydantic-autocompletion-in-vs-code
            # idealy solved by custom templates in the future: https://github.com/koxudaxi/datamodel-code-generator/issues/860

            content = ""
            with open(temp_model_path, "r", encoding="utf-8") as f:
                content = f.read()
            os.remove(temp_model_path)

            # Statement and its subclasses are a complex case that needs manual fixing

            # datamodel-codegen <= 0.15.0
            # make statement classes subclasses of Statement
            content = re.sub(
                r"ObjectStatement\(OswBaseModel\)",
                r"ObjectStatement(Statement)",
                content,
            )
            content = re.sub(
                r"DataStatement\(OswBaseModel\)", r"DataStatement(Statement)", content
            )
            content = re.sub(
                r"QuantityStatement\(OswBaseModel\)",
                r"QuantityStatement(Statement)",
                content,
            )
            # make statement lists union of all statement types
            content = re.sub(
                r"List\[Statement\]",
                r"List[Union[ObjectStatement, DataStatement, QuantityStatement]]",
                content,
            )
            # remove Statement class
            content = re.sub(
                r"(class\s*"
                + "Statement"
                + r"\s*\(\s*\S*\s*\)\s*:.*\n[\s\S]*?(?:[^\S\n]*\n){3,})",
                "",
                content,
                count=1,
            )
            # rename Statement1 to Statement
            content = re.sub(r"Statement1", r"Statement", content)
            # add forward refs
            content = re.sub(
                r"Statement.update_forward_refs\(\)",
                r"Statement.update_forward_refs()\nObjectStatement.update_forward_refs()\nDataStatement.update_forward_refs()\nQuantityStatement.update_forward_refs()",
                content,
            )
            pattern = re.compile(
                r"(class\s*"
                + "Statement"
                + r"\s*\(\s*\S*\s*\)\s*:.*\n[\s\S]*?(?:[^\S\n]*\n){3,})"
            )  # match Statement class definition
            for cls in re.findall(pattern, content):
                # remove class
                content = re.sub(
                    r"(class\s*"
                    + "Statement"
                    + r"\s*\(\s*\S*\s*\)\s*:.*\n[\s\S]*?(?:[^\S\n]*\n){3,})",
                    "",
                    content,
                    count=1,
                )
                content = re.sub(
                    r"(class\s*\S*\s*\(\s*Statement\s*\)\s*:.*\n)",
                    cls + r"\1",
                    content,
                    1,
                )  # insert class definition before first reference
                break

            # datamodel-codegen > 0.15.0
            # Rename statement classes (ObjectStatement, DataStatement, QuantityStatement)
            # content = re.sub(r"ObjectStatement", r"_ObjectStatement", content)
            # content = re.sub(r"DataStatement", r"_DataStatement", content)
            # content = re.sub(r"QuantityStatement", r"_QuantityStatement", content)
            # class Statement1(_ObjectStatement):
            # content = re.sub(r"class\s*\S*(\s*\(\s*_ObjectStatement\s*\))", r"class ObjectStatement\1", content)
            # content = re.sub(r"class\s*\S*(\s*\(\s*_DataStatement\s*\))", r"class DataStatement\1", content)
            # content = re.sub(r"class\s*\S*(\s*\(\s*_QuantityStatement\s*\))", r"class QuantityStatement\1", content)
            # Union[Statement1, Statement2, Statement3] and Statement<x>.update_forward_refs()
            # content = re.sub(r"Statement1", r"ObjectStatement", content)
            # content = re.sub(r"Statement2", r"DataStatement", content)
            # content = re.sub(r"Statement3", r"QuantityStatement", content)

            if fetchSchemaParam.mode == "replace":
                header = (
                    "from uuid import uuid4\n"
                    "from typing import Type, TypeVar\n"
                    "from osw.model.static import OswBaseModel, Ontology\n"
                    "\n"
                )

                content = re.sub(
                    r"(class\s*\S*\s*\(\s*OswBaseModel\s*\)\s*:.*\n)",
                    header + r"\n\n\n\1",
                    content,
                    1,
                )  # add header before first class declaration

                content = re.sub(
                    r"(UUID = Field\(...)",
                    r"UUID = Field(default_factory=uuid4",
                    content,
                )  # enable default value for uuid
                with open(result_model_path, "w", encoding="utf-8") as f:
                    f.write(content)

            if fetchSchemaParam.mode == "append":
                org_content = ""
                with open(result_model_path, "r", encoding="utf-8") as f:
                    org_content = f.read()

                pattern = re.compile(
                    r"class\s*([\S]*)\s*\(\s*\S*\s*\)\s*:.*\n"
                )  # match class definition [\s\S]*(?:[^\S\n]*\n){2,}
                for cls in re.findall(pattern, org_content):
                    print(cls)
                    content = re.sub(
                        r"(class\s*"
                        + cls
                        + r"\s*\(\s*\S*\s*\)\s*:.*\n[\s\S]*?(?:[^\S\n]*\n){3,})",
                        "",
                        content,
                        count=1,
                    )  # replace duplicated classes

                content = re.sub(
                    r"(from __future__ import annotations)", "", content, 1
                )  # remove import statement
                # print(content)
                with open(result_model_path, "a", encoding="utf-8") as f:
                    f.write(content)

            importlib.reload(model)  # reload the updated module
            if not site_cache_state:
                self.site.disable_cache()  # restore original state

    def load_entity(self, entity_title) -> model.Entity:
        """Loads the entity with the given wiki page name from the OSW instance.
        Creates an instance of the class specified by the "type" attribute, default
        model.Entity. An instance of model.Entity can be cast to any subclass with
        .cast(model.<class>) .

        Parameters
        ----------
        entity_title
            the wiki page name

        Returns
        -------
            the dataclass instance
        """
        entity = None
        schemas = []
        page = self.site.get_page(WtSite.GetPageParam(titles=[entity_title])).pages[0]
        jsondata = page.get_slot_content("jsondata")
        if jsondata:
            for category in jsondata["type"]:
                schema = (
                    self.site.get_page(WtSite.GetPageParam(titles=[category]))
                    .pages[0]
                    .get_slot_content("jsonschema")
                )
                schemas.append(schema)

        if len(schemas) == 0:
            print("Error: no schema defined")

        elif len(schemas) == 1:
            cls = schemas[0]["title"]
            entity = eval(f"model.{cls}(**jsondata)")

        else:
            bases = []
            for schema in schemas:
                bases.append(eval("model." + schema["title"]))
            cls = create_model("Test", __base__=tuple(bases))
            entity = cls(**jsondata)

        return entity

    class StoreEntityParam(model.OswBaseModel):
        entities: Union[OswBaseModel, List[OswBaseModel]]
        namespace: Optional[str]
        parallel: Optional[bool] = False
        debug: Optional[bool] = False

    def store_entity(
        self, param: Union[StoreEntityParam, OswBaseModel, List[OswBaseModel]]
    ) -> None:
        """stores the given dataclass instance as OSW page by calling BaseModel.json()

        Parameters
        ----------
        param:
            StoreEntityParam, the dataclass instance or a list of instances
        """
        if isinstance(param, model.Entity):
            param = OSW.StoreEntityParam(entities=[param])
        if isinstance(param, list):
            param = OSW.StoreEntityParam(entities=param)
        if not isinstance(param.entities, list):
            param.entities = [param.entities]

        max_index = len(param.entities)
        if max_index >= 5:
            param.parallel = True

        def store_entity_(
            entity: model.Entity, index: int = None, namespace_=param.namespace
        ) -> None:
            # ToDo: Move this to a separate function entity_to_page()...
            title_ = None
            namespace_ = None
            if hasattr(entity, "meta") and entity.meta and entity.meta.wiki_page:
                if entity.meta.wiki_page.title:
                    title_ = entity.meta.wiki_page.title
                if entity.meta.wiki_page.namespace:
                    namespace_ = entity.meta.wiki_page.namespace
            if namespace_ is None:
                namespace_ = get_namespace(entity)
            if title_ is None:
                title_ = OSW.get_osw_id(entity.uuid)
            if namespace_ is None or title_ is None:
                print("Error: Unsupported entity type")
                return
            entity_title = namespace_ + ":" + title_
            page = self.site.get_page(WtSite.GetPageParam(titles=[entity_title])).pages[
                0
            ]

            jsondata = json.loads(
                entity.json(exclude_none=True)
            )  # use pydantic serialization, skip none values
            page.set_slot_content("jsondata", jsondata)
            page.set_slot_content(
                "header", "{{#invoke:Entity|header}}"
            )  # required for json parsing and header rendering
            page.set_slot_content(
                "footer", "{{#invoke:Entity|footer}}"
            )  # required for footer rendering
            page.edit()
            if index is None:
                print(f"Entity stored at {page.get_url()}.")
            else:
                print(
                    f"({index + 1}/{max_index}) Entity stored at " f"{page.get_url()}."
                )

        if param.parallel:
            _ = parallelize(store_entity_, param.entities, flush_at_end=param.debug)
        else:
            _ = [store_entity_(e, i) for i, e in enumerate(param.entities)]

    class DeleteEntityParam(model.OswBaseModel):
        entities: List[model.OswBaseModel]
        comment: Optional[str] = None
        parallel: Optional[bool] = False
        debug: Optional[bool] = False

    def delete_entity(
        self, entity: Union[model.OswBaseModel, DeleteEntityParam], comment: str = None
    ):
        """Deletes the given entity/entities from the OSW instance."""
        if not isinstance(entity, OSW.DeleteEntityParam):
            if isinstance(entity, list):
                entity = OSW.DeleteEntityParam(entities=entity)
            else:
                entity = OSW.DeleteEntityParam(entities=[entity])
        if comment is not None:
            entity.comment = comment
        if len(entity.entities) >= 5:
            entity.parallel = True

        def delete_entity_(entity, comment_: str = None):
            """Deletes the given entity from the OSW instance.

            Parameters
            ----------
            entity:
                The dataclass instance to delete
            comment_:
                Command for the change log, by default None
            """
            title_ = None
            namespace_ = None
            if hasattr(entity, "meta") and entity.meta and entity.meta.wiki_page:
                if entity.meta.wiki_page.title:
                    title_ = entity.meta.wiki_page.title
                if entity.meta.wiki_page.namespace:
                    namespace_ = entity.meta.wiki_page.namespace
            if namespace_ is None:
                namespace_ = get_namespace(entity)
            if title_ is None:
                title_ = OSW.get_osw_id(entity.uuid)
            if namespace_ is None or title_ is None:
                print("Error: Unsupported entity type")
                return
            entity_title = namespace_ + ":" + title_
            page = self.site.get_page(WtSite.GetPageParam(titles=[entity_title])).pages[
                0
            ]

            if page.exists:
                page.delete(comment_)
                print("Entity deleted: " + page.get_url())
            else:
                print(f"Entity '{entity_title}' does not exist!")

        if entity.parallel:
            _ = parallelize(
                delete_entity_,
                entity.entities,
                flush_at_end=entity.debug,
                comment_=entity.comment,
            )
        else:
            _ = [delete_entity_(e, entity.comment) for e in entity.entities]

    class QueryInstancesParam(model.OswBaseModel):
        categories: List[Union[str, OswBaseModel]]
        parallel: Optional[bool] = False
        debug: Optional[bool] = False
        limit: Optional[int] = 1000

    def query_instances(
        self, category: Union[str, OswBaseModel, OSW.QueryInstancesParam]
    ) -> List[str]:
        def get_page_title(category_: Union[str, OswBaseModel]) -> str:
            error_msg = (
                "Category must be a string, a dataclass instance with "
                "a 'type' attribute or osw.wiki_tools.SearchParam."
            )
            if isinstance(category_, str):
                return category_.split(":")[-1]  # page title w/o namespace
            elif isinstance(category_, model.OswBaseModel):
                type_ = getattr(category_, "type", None)
                if type_:
                    full_page_title = type_[0]
                    return full_page_title.split(":")[-1]  # page title w/o namespace
                else:
                    raise TypeError(error_msg)
            else:
                raise TypeError(error_msg)

        if isinstance(category, OSW.QueryInstancesParam):
            page_titles = [get_page_title(cat) for cat in category.categories]
        else:
            page_titles = [get_page_title(category)]
            category = OSW.QueryInstancesParam(category=page_titles)

        search_param = SearchParam(
            query=[f"[[HasType::Category:{page_title}]]" for page_title in page_titles],
            **category.dict(exclude={"categories"}),
        )
        full_page_titles = self.site.semantic_search(search_param)
        return full_page_titles



from uuid import UUID

from osw.core import OSW
from osw.data.import_utility import get_uuid_from_object_via_type


def get_uuid_from_object_via_type_test():
    expected = UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")
    entity = {"type": "Item:" + OSW.get_osw_id(expected)}
    result = get_uuid_from_object_via_type(entity)
    assert result == expected
